Attribute VB_Name = "basGym2Mid"
Option Explicit

Private Type LowEndianWord
    MSB As Byte
    LSB As Byte
End Type

Private Type LowEndianDoubleWord
    MSB As Byte
    Byte3 As Byte
    Byte2 As Byte
    LSB As Byte
End Type

Private Type ymRegisterData
    LFOEn As Boolean
    LFOFreq As Byte
    TimerA As Integer
    TimerB As Byte
    Ch3Mode As Boolean
    Operators As Byte
    Channel As Byte
    DacEn As Boolean
    DACData As Byte
End Type

Public ymRegData As ymRegisterData

Private Type ymChannelData
    OpNum As Integer
    MidiNoteOn As Boolean
    DT1(3) As Integer
    MUL(3) As Integer
    TL(3) As Integer
    MidiVolume As Integer
    RS(3) As Integer
    AR(3) As Integer
    AM(3) As Integer
    D1R(3) As Integer
    D2R(3) As Integer
    D1L(3) As Integer
    RR(3) As Integer
    FreqMSB As Byte
    FreqLSB As Byte
    Block As Byte
    OldFreq As Long
    Freq As Long
    FreqDifference As Long
    OldMidiNote As Byte
    MidiNote As Byte
    NoteDifference As Integer
    PitchWheel As Integer
    Feedback As Byte
    Algorithm As Byte
    MidiProgram As Byte
    L As Byte
    R As Byte
    OldMidiPan As Byte
    MidiPan As Byte
    AMS As Byte
    FMS As Byte
End Type

Public ymChanData(0 To 8) As ymChannelData

' YM2612 Constants
Private Const YM_SPEAKER_BOTH = 192
Private Const YM_SPEAKER_LEFT = 128
Private Const YM_SPEAKER_RIGHT = 64

' Frequency Conversion Constants
Private Const FREQ_YM = 0
Private Const FREQ_PSG = 1
Private Const FREQ_BASE = 37
Private Const FREQ_CURVE = 1.0594630102 '1.05945917435

Private Type ProgSoundGenData
    NoteOn As Boolean
    FreqMSB As Byte
    FreqLSB As Byte
    OldFreq As Integer
    Freq As Integer
    FreqDifference As Integer
    OldMidiNote As Byte
    MidiNote As Byte
    NoteDifference As Integer
    PitchWheel As Integer
    FB As Byte
    ClockSource As Byte
    Attenuator As Byte
    MidiVolume As Byte
End Type

Public PSGData(0 To 4) As ProgSoundGenData

' PSG Constants
Public Const PSG_TONE_1 = 128
Public Const PSG_TONE_2 = 160
Public Const PSG_TONE_3 = 192
Public Const PSG_NOISE = 224

Public Const MIDI_PROGRAM_SQUARE_WAVE = 80
Public Const PSG_FB_PERIODIC = 0
Public Const MIDI_PROGRAM_PERIODIC = 125
Public Const PSG_FB_WHITE = 1
Public Const MIDI_PROGRAM_WHITE = 126

Public Const PSG_CLOCK_SOURCE_HALF = 0
Public Const PSG_CLOCK_SOURCE_FOURTH = 1
Public Const PSG_CLOCK_SOURCE_EIGHTH = 2
Public Const PSG_CLOCK_SOURCE_TONE_3 = 3

Public Const PSG_ATTENUATOR_1 = 144
Public Const PSG_ATTENUATOR_2 = 176
Public Const PSG_ATTENUATOR_3 = 208
Public Const PSG_ATTENUATOR_NOISE = 240
Public Const PSG_ATTENUATION_MIN = 0
Public Const PSG_ATTENUATION_MAX = 15

' Variables for storing and reading .GYM data
Public GymData() As Byte
Public GymFilePos As Long

Private Type MidiHeaderData
    MThd As String * 4
    HeaderLen As LowEndianDoubleWord
    MidiFormat As LowEndianWord
    NumTracks As LowEndianWord
    NumTicks As LowEndianWord
End Type

Public MidiHeader As MidiHeaderData

Private Type MidiTrackHeaderData
    Filler(1 To 14) As Byte
    MTrk As String * 4
    TrackLen As LowEndianDoubleWord
End Type

Public MidiTrackHeader As MidiTrackHeaderData

Public MidData() As Byte
Public MidFilePos As Long

' MIDI Command Constants
Public Const MIDI_NOTE_OFF = &H80
Public Const MIDI_NOTE_ON = &H90
Public Const MIDI_NOTE_AFTERTOUCH = &HA0
Public Const MIDI_CONTROLLER_CHANGE = &HB0
Public Const MIDI_PROGRAM_CHANGE = &HC0
Public Const MIDI_CHANNEL_AFTERTOUCH = &HD0
Public Const MIDI_PITCH_WHEEL = &HE0
Public Const META_EVENT = &HFF

' MIDI Controller Numbers
Public Const MIDI_DATA_ENTRY = 6
Public Const MIDI_VOLUME = 7
Public Const MIDI_PAN = 10
Public Const MIDI_LEGATO_PEDAL = 68
Public Const MIDI_HOLD2_PEDAL = 69
Public Const MIDI_SOUND_TIMBRE = 71
Public Const MIDI_SOUND_RELEASE_TIME = 72
Public Const MIDI_SOUND_ATTACK_TIME = 73
Public Const MIDI_SOUND_BRIGHTNESS = 74
Public Const MIDI_RPN_FINE = 100
Public Const MIDI_RPN_COARSE = 101
Public Const MIDI_ALL_CONTROLLERS_OFF = 123
Public Const MIDI_ALL_NOTES_OFF = 123

' MIDI Controller Values
Public Const MIDI_VOLUME_MIN = &H0
Public Const MIDI_VOLUME_MAX = &H7F

Public Const MIDI_PAN_LEFT = 31
Public Const MIDI_PAN_RIGHT = 95
Public Const MIDI_PAN_CENTER = 63

Public Const RPN_PITCH_BEND_RANGE = 0
Public Const PITCH_BEND_CHANGE = 64

' Meta Event Constants
Public Const META_TRACK_END = &H2F

' Variables for gym2mid options
Public StereoSound As Boolean
Public PatchChanges As Boolean
Public VolumeLevels As Boolean
Public OldPitchBend As Boolean
Private Function CYm28hToMidiNoteOn(ByRef lngDelay As Long, ByVal chrPort As Byte, ByVal chrReg As Byte, ByVal chrData As Byte) As Long
        
        With ymRegData
            .Channel = chrData And 7
            .Channel = IIf(.Channel > 3, .Channel - 1, .Channel)
            .Operators = (chrData And 240) / 16
        End With
        With ymChanData(ymRegData.Channel)
            If .MidiNoteOn = True Then
                WriteMidiEvent lngDelay, MIDI_NOTE_OFF, ymRegData.Channel, .MidiNote, &H0
                lngDelay = 0
            End If
                If .PitchWheel <> &H2000 Then
                    .PitchWheel = &H2000
                    WriteMidiEvent lngDelay, MIDI_PITCH_WHEEL, ymRegData.Channel, .PitchWheel
                    lngDelay = 0
                End If
        End With
        Select Case ymRegData.Operators
        Case 0
            With ymChanData(ymRegData.Channel)
                .MidiNoteOn = False
            End With
        Case Else
            With ymChanData(ymRegData.Channel)
                .OldMidiNote = .MidiNote
                .MidiNote = CFreqToMidiNote(.Freq, FREQ_YM)
                WriteMidiEvent lngDelay, MIDI_NOTE_ON, ymRegData.Channel, .MidiNote, &H7F
                .MidiNoteOn = True
                lngDelay = 0
            End With
        End Select

    CYm28hToMidiNoteOn = lngDelay

End Function
Private Function CFreqToMidiNote(ByVal Freq As Long, ByVal FreqType As Byte) As Byte
    Dim Counter As Double, NewMidiNote As Integer
    
    Counter = FREQ_BASE
    NewMidiNote = -1
    
    Do
        NewMidiNote = NewMidiNote + 1
        Counter = Counter * FREQ_CURVE
    Loop Until Counter >= Freq
        
    NewMidiNote = IIf(FreqType = FREQ_YM, NewMidiNote - 37, 102 - NewMidiNote)
    
    If NewMidiNote < 0 Then
        NewMidiNote = 0
    ElseIf NewMidiNote > 127 Then
        NewMidiNote = 127
    End If
        
    CFreqToMidiNote = NewMidiNote
End Function
Private Function CYm4xhToMidiVolume(ByRef lngDelay As Long, ByVal chrPort As Byte, ByVal chrReg As Byte, ByVal chrData As Byte) As Long

        ymRegData.Channel = (((chrReg Xor &H40) / 12) - Fix((chrReg Xor &H40) / 12)) + chrPort
        With ymChanData(ymRegData.Channel)
            .OpNum = (Fix((chrReg Xor &H40) / 4))
            .TL(.OpNum) = chrData And 127
            Select Case .Algorithm
            Case 0, 1, 2, 3:
                Select Case .OpNum
                Case 3
                    .MidiVolume = .TL(3)
                Case Else
                    .MidiVolume = -1
                End Select
            Case 4:
                Select Case .OpNum
                Case 1, 3
'                    .MidiVolume = .TL(1)
'                    .MidiVolume = (.MidiVolume + .TL(3)) / 2
                    .MidiVolume = .TL(1) And .TL(3)
                Case Else
                    .MidiVolume = -1
                End Select
            Case 5, 6:
                Select Case .OpNum
                Case 1, 2, 3
'                    .MidiVolume = .TL(1)
'                    .MidiVolume = .MidiVolume + .TL(2)
'                    .MidiVolume = (.MidiVolume + .TL(3)) / 3
                    .MidiVolume = .TL(1) And .TL(2) And .TL(3)
                Case Else
                    .MidiVolume = -1
                End Select
            Case 7:
                Select Case .OpNum
                Case 0, 1, 2, 3
'                    .MidiVolume = .TL(0)
'                    .MidiVolume = .MidiVolume + .TL(1)
'                    .MidiVolume = .MidiVolume + .TL(2)
'                    .MidiVolume = (.MidiVolume + .TL(3)) / 4
                    .MidiVolume = .TL(0) And .TL(1) And .TL(2) And .TL(3)
                Case Else
                    .MidiVolume = -1
                End Select
            End Select
            If .MidiVolume <> -1 Then
                .MidiVolume = (MIDI_VOLUME_MAX - .MidiVolume) * 0.75
                WriteMidiEvent lngDelay, MIDI_CONTROLLER_CHANGE, ymRegData.Channel, MIDI_VOLUME, .MidiVolume
                lngDelay = 0
            End If
        End With

    CYm4xhToMidiVolume = lngDelay

End Function
Private Function CYmAxhToMidiPitchWheel(ByRef lngDelay As Long, ByVal chrPort As Byte, ByVal chrReg As Byte, ByVal chrData As Byte) As Long
    
    With ymChanData(ymRegData.Channel)
        Select Case OldPitchBend
        Case True
            .OldMidiNote = .MidiNote
            .MidiNote = CFreqToMidiNote(.Freq, FREQ_YM)
            If .MidiNote <> .OldMidiNote Then
                WriteMidiEvent lngDelay, MIDI_NOTE_OFF, ymRegData.Channel, .OldMidiNote, 0
                WriteMidiEvent 0, MIDI_NOTE_ON, ymRegData.Channel, .MidiNote, &H7F
                .MidiNoteOn = True
                lngDelay = 0
            End If
        Case False
            .FreqDifference = .Freq
            .FreqDifference = .FreqDifference - .OldFreq
            If .FreqDifference <> 0 Then
                .OldMidiNote = .MidiNote
                .MidiNote = CFreqToMidiNote(.Freq, FREQ_YM)
                If .MidiNote <> .OldMidiNote Then
                    .NoteDifference = .MidiNote
                    .NoteDifference = .NoteDifference - .OldMidiNote
                    .MidiNote = .OldMidiNote
                    .PitchWheel = &H2000 + (.NoteDifference * (128))
                    WriteMidiEvent lngDelay, MIDI_PITCH_WHEEL, ymRegData.Channel, Fix(.PitchWheel)
                    lngDelay = 0
                End If
            End If
        End Select
    End With

    CYmAxhToMidiPitchWheel = lngDelay

End Function
Private Function CYmB0hToMidiProgram(ByRef lngDelay As Long, ByVal chrPort As Byte, ByVal chrReg As Byte, ByVal chrData As Byte) As Long
        
    ymRegData.Channel = (chrReg Xor &HB0) + chrPort
    With ymChanData(ymRegData.Channel)
        .Algorithm = chrData And 7
        .MidiProgram = 87
        WriteMidiEvent lngDelay, MIDI_PROGRAM_CHANGE, ymRegData.Channel, .MidiProgram
        lngDelay = 0
    End With

    CYmB0hToMidiProgram = lngDelay

End Function
Private Function CYmB4hToMidiPan(ByRef lngDelay As Long, ByVal chrPort As Byte, ByVal chrReg As Byte, ByVal chrData As Byte) As Long
    
    ymRegData.Channel = (chrReg Xor &HB4) + chrPort
    With ymChanData(ymRegData.Channel)
        .L = chrData And 128
        .R = chrData And 64
        .OldMidiPan = .MidiPan
        Select Case (.L + .R)
        Case YM_SPEAKER_BOTH
            .MidiPan = MIDI_PAN_CENTER
        Case YM_SPEAKER_LEFT
            .MidiPan = MIDI_PAN_LEFT
        Case YM_SPEAKER_RIGHT
            .MidiPan = MIDI_PAN_RIGHT
        End Select
        ymRegData.Channel = IIf(ymRegData.Channel = 5 And ymRegData.DacEn = True, 9, ymRegData.Channel)
        If .MidiPan <> .OldMidiPan Then
            WriteMidiEvent lngDelay, MIDI_CONTROLLER_CHANGE, ymRegData.Channel, MIDI_PAN, .MidiPan
            lngDelay = 0
        End If
    End With
    
    If ymRegData.Channel = 9 Then
        WriteMidiEvent lngDelay, MIDI_NOTE_OFF, ymRegData.Channel, 42, &H0
        WriteMidiEvent 0, MIDI_NOTE_ON, ymRegData.Channel, 42, &H7F
        lngDelay = 0
    End If

    CYmB4hToMidiPan = lngDelay

End Function
Private Function Initialize_Data()

    Dim CounterA As Byte, CounterB As Byte

    With ymRegData
        .LFOEn = False
        .LFOFreq = 0
        .TimerA = 0
        .TimerB = 0
        .Ch3Mode = False
        .Operators = 0
        .Channel = 0
        .DacEn = False
        .DACData = 0
    End With
    
    For CounterA = 0 To UBound(ymChanData)
        With ymChanData(CounterA)
            .OpNum = 0
            .MidiNoteOn = False
            For CounterB = 0 To 3
                .DT1(CounterB) = 0
                .MUL(CounterB) = 0
                .TL(CounterB) = 0
                .RS(CounterB) = 0
                .AR(CounterB) = 0
                .AM(CounterB) = 0
                .D1R(CounterB) = 0
                .D2R(CounterB) = 0
                .D1L(CounterB) = 0
                .RR(CounterB) = 0
            Next
            .MidiVolume = 0
            .FreqMSB = 0
            .FreqLSB = 0
            .Block = 0
            .OldFreq = 0
            .Freq = 0
            .FreqDifference = 0
            .OldMidiNote = 0
            .MidiNote = 0
            .NoteDifference = 0
            .PitchWheel = 0
            .Feedback = 0
            .Algorithm = 0
            .MidiProgram = 0
            .L = 0
            .R = 0
            .OldMidiPan = 0
            .MidiPan = 0
            .AMS = 0
            .FMS = 0
        End With
    Next
   
    For CounterA = 0 To UBound(PSGData)
        With PSGData(CounterA)
            .NoteOn = False
            .FreqMSB = 0
            .FreqLSB = 0
            .OldFreq = 0
            .Freq = 0
            .FreqDifference = 0
            .OldMidiNote = 0
            .MidiNote = 0
            .NoteDifference = 0
            .FB = 0
            .ClockSource = 0
            .Attenuator = 0
            .MidiVolume = 0
        End With
    Next

    ' Initialize .GYM data
    Erase GymData
    ReDim GymData(0) As Byte
    GymFilePos = 0

    ' Initialize .MID data
    Erase MidData
    ReDim MidData(0) As Byte
    MidFilePos = 0

    ' Pan (Stereo) Settings for YM Channels 1-9 (Ch 3 Special)
    For CounterA = 0 To 8
        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_PAN, MIDI_PAN_CENTER
    Next
    
    ' Program Changes for YM Channels 1-9 (Ch 3 Special)
    For CounterA = 0 To 8
        WriteMidiEvent 0, MIDI_PROGRAM_CHANGE, CounterA, 87
    Next

    ' Initial Volume Levels for YM Channels 1-9 (Ch 3 Special)
    For CounterA = 0 To 8
        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_VOLUME, 112
    Next

    ' Change Pitch Bend Range for YM Channels 1-9 (Ch 3 Special)
    For CounterA = 0 To 8
        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_RPN_COARSE, RPN_PITCH_BEND_RANGE
        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_DATA_ENTRY, PITCH_BEND_CHANGE
    Next

    ' Pan (Stereo) Settings for PSG Channels 1-3 and Noise Channel
    For CounterA = 10 To 13
        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_PAN, MIDI_PAN_CENTER
    Next
    
    ' Program Changes for PSG Channels 1-3
    For CounterA = 10 To 12
        WriteMidiEvent 0, MIDI_PROGRAM_CHANGE, CounterA, MIDI_PROGRAM_SQUARE_WAVE
    Next

    ' Initial Volume Levels for PSG Channels 1-3 and Noise Channel
    For CounterA = 10 To 13
        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_VOLUME, 127
    Next

'    ' Change Pitch Bend Range for PSG Channels 1-3 and Noise Channel
'    For CounterA = 10 To 13
'        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_RPN_COARSE, RPN_PITCH_BEND_RANGE
'        WriteMidiEvent 0, MIDI_CONTROLLER_CHANGE, CounterA, MIDI_DATA_ENTRY, 2
'    Next

End Function
Private Function CIntToLEWord(intInt As Integer) As LowEndianWord
    Dim intMSB As Integer, intLSB As Integer
    With CIntToLEWord
        intMSB = Fix((intInt / 256)) And 255
        .MSB = CByte(intMSB)
        intLSB = intInt And 255
        .LSB = CByte(intLSB)
    End With
End Function
Private Function CIntToMidiPitchWheel(intInt As Integer) As LowEndianWord
    Dim intMSB As Integer, intLSB As Integer
    With CIntToMidiPitchWheel
        intMSB = Fix((intInt / 128)) And 127
        .MSB = CByte(intMSB)
        intLSB = intInt And 127
        .LSB = CByte(intLSB)
    End With
End Function


Private Function CLEDblWordToLng(LEDblWord As LowEndianDoubleWord) As Long
    With LEDblWord
        CLEDblWordToLng = .LSB + (.Byte2 * 256) + (.Byte3 * (256 ^ 2)) + (.MSB * (256 ^ 3))
    End With
End Function
Private Function CLEWordToInt(LEWord As LowEndianWord) As Integer
    With LEWord
        CLEWordToInt = .LSB + (.MSB * 256)
    End With
End Function
Private Function CLngToLEDblWord(lngLong As Long) As LowEndianDoubleWord
    Dim lngMSB As Long, lngByte3 As Long, intByte2 As Integer, intLSB As Integer
    With CLngToLEDblWord
        lngMSB = Fix((lngLong / (256 ^ 3))) And 255
        .MSB = CByte(lngMSB)
        lngByte3 = Fix((lngLong / (256 ^ 2))) And 255
        .Byte3 = CByte(lngByte3)
        intByte2 = Fix((lngLong / 256)) And 255
        .Byte2 = CByte(intByte2)
        intLSB = lngLong And 255
        .LSB = CByte(intLSB)
    End With
End Function
Public Function Convert_GYM_Data(ByRef lngDelay As Long, ByVal chrPort As Byte, ByVal chrReg As Byte, ByVal chrData As Byte) As Long
        
    Select Case chrReg
'    Case &H22:
'        ymRegData.LFOEn = (chrData And 8) / 8
'        ymRegData.LFOFreq = chrData And 7
'    Case &H24:
'        ymRegData.TimerA = 0
'        ymRegData.TimerA = chrData
'    Case &H25:
'        ymRegData.TimerA = ymRegData.TimerA + (chrData * 256)
'    Case &H27:
'        Select Case (chrData And 192) / 64
'        Case 0:
'            ymRegData.Ch3Mode = False
'        Case 1:
'            ymRegData.Ch3Mode = True
'        End Select
    Case &H28:
        lngDelay = CYm28hToMidiNoteOn(lngDelay, chrPort, chrReg, chrData)
'    Case &H2A:
'        ymRegData.DACData = chrData
    Case &H2B:
        Select Case (chrData And 128) / 128
        Case 0:
            ymRegData.DacEn = False
        Case 1:
            ymRegData.DacEn = True
        End Select
'    Case &H30 To &H3E:
'        ymRegData.Channel = (((chrReg Xor &H30) / 12) - Fix((chrReg Xor &H30) / 12)) + chrPort
'        With ymChanData(ymRegData.Channel)
'            .OpNum = (Fix((chrReg Xor &H30) / 4))
'            .DT1(.OpNum) = (chrData And 240) / 16
'            .MUL(.OpNum) = chrData And 15
'            Debug.Print ymRegData.Channel; ": "; .MUL(0) Or .MUL(1) Or .MUL(2) Or .MUL(3)
'        End With
    Case &H40 To &H4E:
        If VolumeLevels = True Then lngDelay = CYm4xhToMidiVolume(lngDelay, chrPort, chrReg, chrData)
'        Case &H50 To &H5E:
'            ymRegData.Channel = (((chrReg Xor &H50) / 12) - Fix((chrReg Xor &H50) / 12)) + chrPort
'            With ymChanData(ymRegData.Channel)
'                .OpNum = (Fix((chrReg Xor &H50) / 4))
'                .RS(.OpNum) = (chrData And 192) / 192
'                .AR(.OpNum) = chrData And 15
'            End With
'        Case &H60 To &H6E:
'            ymRegData.Channel = (((chrReg Xor &H60) / 12) - Fix((chrReg Xor &H60) / 12)) + chrPort
'            With ymChanData(ymRegData.Channel)
'                .OpNum = (Fix((chrReg Xor &H60) / 4))
'                .AM(.OpNum) = (chrData And 128) / 128
'                .D1R(.OpNum) = chrData And 31
'            End With
'        Case &H70 To &H7E:
'            ymRegData.Channel = (((chrReg Xor &H70) / 12) - Fix((chrReg Xor &H70) / 12)) + chrPort
'            With ymChanData(ymRegData.Channel)
'                .OpNum = (Fix((chrReg Xor &H70) / 4))
'                .D2R(.OpNum) = chrData And 31
'            End With
'        Case &H80 To &H8E:
'            ymRegData.Channel = (((chrReg Xor &H80) / 12) - Fix((chrReg Xor &H80) / 12)) + chrPort
'            With ymChanData(ymRegData.Channel)
'                .OpNum = (Fix((chrReg Xor &H80) / 4))
'                .D1L(.OpNum) = chrData And 240
'                .RR(.OpNum) = chrData And 15
'            End With
    Case &HA0 To &HAF:
        Select Case chrReg:
        Case &HA4 To &HA6
            ymRegData.Channel = (chrReg Xor &HA4) + chrPort
            With ymChanData(ymRegData.Channel)
                .FreqMSB = chrData
            End With
        Case &HA0 To &HA2
            ymRegData.Channel = (chrReg Xor &HA0) + chrPort
            With ymChanData(ymRegData.Channel)
                .Block = (.FreqMSB And 56) / 8
                .FreqLSB = chrData
                .OldFreq = .Freq
                .Freq = ((.FreqMSB And 7) * 256) + .FreqLSB
                .Freq = IIf(.Block = 0, .Freq, .Freq * (2 ^ .Block))
                If .MidiNoteOn = True Then lngDelay = CYmAxhToMidiPitchWheel(lngDelay, chrPort, chrReg, chrData)
            End With
        End Select
    Case &HB0 To &HB2:
'        lngDelay = CYmB0hToMidiProgram(lngDelay, chrPort, chrReg, chrData)
        If PatchChanges = True Then lngDelay = CYmB0hToMidiProgram(lngDelay, chrPort, chrReg, chrData)
    Case &HB4 To &HB6:
'        lngDelay = CYmB4hToMidiPan(lngDelay, chrPort, chrReg, chrData)
        If StereoSound = True Then lngDelay = CYmB4hToMidiPan(lngDelay, chrPort, chrReg, chrData)
'    Case Else
'        Debug.Print Hex(chrReg)
    End Select

    Convert_GYM_Data = lngDelay
End Function
Public Function Convert_GYM_to_MID(ByVal strFilename As String)

    Initialize_Data
    
'    frmMain.lblConvProg.Caption = "reading .gym data..."
    frmConvProg.lblConvProg.Caption = "reading .gym data..."
    DoEvents
    
    Read_GYM_Data (strFilename)
    
'    frmMain.lblConvProg.Caption = "converting .gym data..."
    frmConvProg.lblConvProg.Caption = "converting .gym data..."
    DoEvents
    
    Interpret_GYM_Data
    
'    frmMain.lblConvProg.Caption = "writing .mid data..."
    frmConvProg.lblConvProg.Caption = "writing .mid data..."
    DoEvents
    
End Function
Private Function Convert_PSG_Data(ByRef lngDelay As Long, ByVal chrPort As Byte, ByVal chrMSB As Byte, ByVal chrLSB As Byte) As Long

    Dim psgChannel As Byte
    
    Select Case chrMSB And 240
    Case PSG_TONE_1, PSG_TONE_2, PSG_TONE_3
        chrLSB = GymData(GymFilePos + 3)
        psgChannel = (chrMSB And 112) / 32
        psgChannel = IIf(psgChannel = 2 And PSGData(3).ClockSource = PSG_CLOCK_SOURCE_TONE_3, 3, psgChannel)
        With PSGData(psgChannel)
            .FreqMSB = chrMSB And 15
            .FreqLSB = chrLSB And 63
            .OldFreq = .Freq
            .Freq = (.FreqLSB * 16) + .FreqMSB
            Select Case .Freq
            Case 0
                If .NoteOn = True Then
                    WriteMidiEvent lngDelay, MIDI_NOTE_OFF, psgChannel + chrPort, .MidiNote, &H0
                    .NoteOn = False
                    lngDelay = 0
                End If
            Case Else
                .OldMidiNote = .MidiNote
                .MidiNote = CFreqToMidiNote(.Freq, FREQ_PSG)
                If .NoteOn = False Or .MidiNote <> .OldMidiNote Then
                    WriteMidiEvent lngDelay, MIDI_NOTE_OFF, psgChannel + chrPort, .OldMidiNote, &H0
                    WriteMidiEvent 0, MIDI_NOTE_ON, psgChannel + chrPort, .MidiNote, &H7F
                    .NoteOn = True
                    lngDelay = 0
                End If
            End Select
        End With
        GymFilePos = GymFilePos + 3
    Case PSG_NOISE
        psgChannel = 3
        With PSGData(psgChannel)
            .FB = chrMSB And 4 / 4
            .ClockSource = chrMSB And 3
            Select Case .FB
            Case PSG_FB_PERIODIC
                WriteMidiEvent lngDelay, MIDI_PROGRAM_CHANGE, psgChannel + chrPort, MIDI_PROGRAM_PERIODIC
                lngDelay = 0
            Case PSG_FB_WHITE
                WriteMidiEvent lngDelay, MIDI_PROGRAM_CHANGE, psgChannel + chrPort, MIDI_PROGRAM_WHITE
                lngDelay = 0
            End Select
            Select Case .ClockSource
            Case PSG_CLOCK_SOURCE_HALF
                .OldMidiNote = .MidiNote
                .MidiNote = 95
            Case PSG_CLOCK_SOURCE_FOURTH
                .OldMidiNote = .MidiNote
                .MidiNote = 60
            Case PSG_CLOCK_SOURCE_EIGHTH
                .OldMidiNote = .MidiNote
                .MidiNote = 31
            Case PSG_CLOCK_SOURCE_TONE_3
                .OldMidiNote = .MidiNote
                .MidiNote = 60
            End Select
            WriteMidiEvent lngDelay, MIDI_NOTE_OFF, psgChannel + chrPort, .OldMidiNote, &H0
            WriteMidiEvent 0, MIDI_NOTE_ON, psgChannel + chrPort, .MidiNote, &H7F
            lngDelay = 0
        End With
        GymFilePos = GymFilePos + 1
    Case PSG_ATTENUATOR_1, PSG_ATTENUATOR_2, PSG_ATTENUATOR_3, PSG_ATTENUATOR_NOISE
        psgChannel = (((chrMSB And 112) / 16) - 1) / 2
        With PSGData(psgChannel)
            .Attenuator = (chrMSB And 15) * 8.45
'            If .Attenuator = 120 Then .Attenuator = 127
            .MidiVolume = MIDI_VOLUME_MAX - .Attenuator
            WriteMidiEvent lngDelay, MIDI_CONTROLLER_CHANGE, psgChannel + chrPort, MIDI_VOLUME, .MidiVolume
            lngDelay = 0
        End With
        GymFilePos = GymFilePos + 1
    Case Else
        GymFilePos = GymFilePos + 1
    End Select

    Convert_PSG_Data = lngDelay

End Function
Private Function Interpret_GYM_Data()
    
    Dim Switch As Byte
    Dim ymDelay As Long, ymPort As Byte, ymReg As Byte, ymData As Byte
    Dim psgPort As Byte, psgMSB As Byte, psgLSB As Byte
    
    ' Interpret .GYM file format
    For GymFilePos = 0 To UBound(GymData)
        Switch = GymData(GymFilePos)
        Select Case Switch
        Case 0
            ymDelay = ymDelay + 1
        Case 1, 2
            ymPort = IIf(Switch = 1, 0, 3)
            ymReg = GymData(GymFilePos + 1)
            ymData = GymData(GymFilePos + 2)
            GymFilePos = GymFilePos + 2
            ymDelay = Convert_GYM_Data(ymDelay, ymPort, ymReg, ymData)
        Case 3
            psgPort = 10
            psgMSB = GymData(GymFilePos + 1)
            ymDelay = Convert_PSG_Data(ymDelay, psgPort, psgMSB, psgLSB)
        End Select
        ymPort = 0
        ymReg = 0
        ymData = 0
        psgPort = 0
        psgMSB = 0
        psgLSB = 0
    Next
End Function
Private Function Read_GYM_Data(ByVal strFilename As String)

    ' Read data from .GYM file
    Open strFilename For Binary As #1
        ReDim GymData(LOF(1)) As Byte
        GymData = InputB(LOF(1), 1)
    Close

End Function
Public Function WriteMid(MidiFileName As String)
    Dim lngMThd As Long, lngHeaderLen As Long, intMidiFormat As Integer, intNumTracks As Integer, intNumTicks As Integer, lngTrackLen As Long
    
    ' Track End marker
    WriteMidiEvent 0, META_EVENT, META_TRACK_END, 0
    
    ' Prepare .MID track header
    With MidiTrackHeader
        .MTrk = "MTrk"
        lngTrackLen = MidFilePos + 1
        .TrackLen = CLngToLEDblWord(lngTrackLen)
    End With
    
    If Dir(MidiFileName) <> "" Then Kill (MidiFileName)
    
    Open MidiFileName For Random As #1 Len = Len(MidiTrackHeader)
        Put #1, 1, MidiTrackHeader
    Close
    
    With MidiHeader
        .MThd = "MThd"
        lngHeaderLen = 6
        .HeaderLen = CLngToLEDblWord(lngHeaderLen)
        intMidiFormat = 0
        .MidiFormat = CIntToLEWord(intMidiFormat)
        intNumTracks = 1
        .NumTracks = CIntToLEWord(intNumTracks)
        intNumTicks = 30
        .NumTicks = CIntToLEWord(intNumTicks)
    End With
    
    Open MidiFileName For Random As #1 Len = Len(MidiHeader)
        Put #1, 1, MidiHeader
    Close
    
    Open MidiFileName For Binary As #1
        Put #1, 23, MidData
    Close
    
End Function
Private Function WriteMidiEvent(ByVal DeltaTime As Long, Optional ByVal Param1 As Variant, Optional ByVal Param2 As Variant, Optional ByVal Param3 As Variant, Optional ByVal Param4 As Variant, Optional ByVal Param5 As Variant) As Byte
    
    If DeltaTime >= 128 Then
        Do Until DeltaTime < 128
            ReDim Preserve MidData(MidFilePos) As Byte
            MidData(MidFilePos) = 128
            MidFilePos = MidFilePos + 1
            DeltaTime = DeltaTime - 128
        Loop
    End If

    Dim Command As Byte, Channel As Byte
    Command = CByte(Param1)

    Select Case Command
    Case MIDI_NOTE_OFF, MIDI_NOTE_ON:
        Dim Note As Byte, Velocity As Byte
        Channel = CByte(Param2)
        Note = CByte(Param3)
        Velocity = CByte(Param4)
        
        ReDim Preserve MidData(MidFilePos + 3) As Byte
        MidData(MidFilePos) = DeltaTime
        MidData(MidFilePos + 1) = (Command Or Channel)
        MidData(MidFilePos + 2) = Note
        MidData(MidFilePos + 3) = Velocity
        MidFilePos = MidFilePos + 4
    Case MIDI_CONTROLLER_CHANGE:
        Dim Controller As Byte, ControllerValue As Byte
        Channel = CByte(Param2)
        Controller = CByte(Param3)
        ControllerValue = CByte(Param4)
        
        ReDim Preserve MidData(MidFilePos + 3) As Byte
        MidData(MidFilePos) = DeltaTime
        MidData(MidFilePos + 1) = (Command Or Channel)
        MidData(MidFilePos + 2) = Controller
        MidData(MidFilePos + 3) = ControllerValue
        MidFilePos = MidFilePos + 4
    Case MIDI_PROGRAM_CHANGE:
        Dim Program As Byte
        Channel = CByte(Param2)
        Program = CByte(Param3)
                
        ReDim Preserve MidData(MidFilePos + 2) As Byte
        MidData(MidFilePos) = DeltaTime
        MidData(MidFilePos + 1) = (Command Or Channel)
        MidData(MidFilePos + 2) = Program
        MidFilePos = MidFilePos + 3
    Case MIDI_PITCH_WHEEL:
        Dim PitchWheel As Integer
        Channel = CByte(Param2)
        PitchWheel = CInt(Param3)
        
        Dim lewPitchWheel As LowEndianWord
        lewPitchWheel = CIntToMidiPitchWheel(PitchWheel)
        
        ReDim Preserve MidData(MidFilePos + 3) As Byte
        MidData(MidFilePos) = DeltaTime
        MidData(MidFilePos + 1) = (Command Or Channel)
        MidData(MidFilePos + 2) = lewPitchWheel.LSB
        MidData(MidFilePos + 3) = lewPitchWheel.MSB
        MidFilePos = MidFilePos + 4
    Case META_EVENT:
        Dim EventType As Byte
        EventType = CByte(Param2)
        Select Case EventType
        Case META_TRACK_END:
            Dim TrackEndMarker As Byte
            TrackEndMarker = CByte(Param3)
            
            ReDim Preserve MidData(MidFilePos + 3) As Byte
            MidData(MidFilePos) = DeltaTime
            MidData(MidFilePos + 1) = Command
            MidData(MidFilePos + 2) = EventType
            MidData(MidFilePos + 3) = TrackEndMarker
            MidFilePos = MidFilePos + 3
        End Select
    End Select

End Function
